use crate::gsa_render_to_screen::render_to_window;
use crate::state::State;
use crate::tileset::load_tileset;
use crate::{Gsa, Rgb, Sprite, MAX_SPRITES, SCREEN_HEIGHT, SCREEN_WIDTH};
use gilrs::Gilrs;
use glam::IVec2;
use std::cmp::min;
use std::num::NonZeroU32;
use std::time::{Duration, Instant};
use winit::dpi::LogicalSize;
use winit::event::{Event, VirtualKeyCode, WindowEvent};
use winit::event_loop::{ControlFlow, EventLoop};
use winit::window::WindowBuilder;

/// Creates main function, includes gfx.gif, and calls run
///
/// Pass the following two functions to it:
/// - fn init(gsa: &mut Gsa) -> TGame
/// - fn update(game: &mut Game, gsa: &mut Gsa)
///
/// TGame can be any type, usually a struct that contains your game's state
#[macro_export]
macro_rules! run {
    ($init: ident, $update: ident) => {
        fn main() {
            run($init, $update, include_bytes!("gfx.gif"));
        }
    };
}

/// This is called by [run!]
pub fn run<TGame: 'static>(
    init_fn: fn(gsa: &mut Gsa) -> TGame,
    update_fn: fn(game: &mut TGame, gsa: &mut Gsa),
    image_data: &[u8],
) {
    let (tileset, palette) = load_tileset(image_data);

    let mut gsa = Gsa {
        sprites: [Sprite::default(); MAX_SPRITES],
        maps: Default::default(),
        palette,
        font: 0x1010,
        input: Default::default(),
    };
    let mut game = init_fn(&mut gsa);
    /*let state = State::<TGame> {
        gsa,
        game,
        tileset,
        update_fn,
        first: true,
        gilrs: Gilrs::new().unwrap(),
    };

     */

    let event_loop = EventLoop::new();
    let size = LogicalSize::new(SCREEN_WIDTH as u32, SCREEN_HEIGHT as u32);
    let window = WindowBuilder::new()
        .with_title("Game Skunk Advance v0.1")
        .with_inner_size(size)
        .with_min_inner_size(size)
        .build(&event_loop)
        .unwrap();

    let context = unsafe { softbuffer::Context::new(&window) }.unwrap();
    let mut surface = unsafe { softbuffer::Surface::new(&context, &window) }.unwrap();

    let mut frames_since_last_second = 0;
    let mut last_second = Instant::now();

    let mut last_frame = last_second;
    let target_frame_duration = Duration::from_secs(1).div_f64(6000.0);

    let mut scale = 1usize;
    let mut off_x = 0usize;
    let mut off_y = 0usize;

    event_loop.run(move |event, _, control_flow| {
        let scale = &mut scale;
        let off_x = &mut off_x;
        let off_y = &mut off_y;
        if Instant::now() - last_second > Duration::from_secs(1) {
            println!("fps: {}", frames_since_last_second);
            frames_since_last_second = 0;
            last_second += Duration::from_secs(1);
        }
        match event {
            Event::WindowEvent {
                event: WindowEvent::KeyboardInput { input, .. },
                ..
            } => match input.virtual_keycode {
                Some(VirtualKeyCode::Escape) => {
                    *control_flow = ControlFlow::Exit;
                }
                _ => {}
            },
            Event::MainEventsCleared {} => {
                if Instant::now() - last_frame >= target_frame_duration {
                    update_fn(&mut game, &mut gsa);
                    let size = window.inner_size();
                    surface
                        .resize(
                            NonZeroU32::new(size.width).unwrap(),
                            NonZeroU32::new(size.height).unwrap(),
                        )
                        .unwrap();

                    *scale = min(
                        size.width / SCREEN_WIDTH as u32,
                        size.height / SCREEN_HEIGHT as u32,
                    ) as usize;
                    *off_x = (size.width as usize - SCREEN_WIDTH * *scale) / 2;
                    *off_y = (size.height as usize - SCREEN_HEIGHT * *scale) / 2;

                    let mut buf = surface.buffer_mut().unwrap();
                    render_to_window(
                        &mut buf,
                        &gsa,
                        IVec2 {
                            x: size.width as i32,
                            y: size.height as i32,
                        },
                    );
                    buf.present().unwrap();
                    frames_since_last_second += 1;
                    last_frame += target_frame_duration;
                }
            }
            _ => {}
        }
    });
}
